@@ -1,18 +1,19 @@
 """
-position_paradigm is the most important module in cgt. PositionParadigmFramework keeps track of important objects, can 
-convert group elements and genomes between different forms, and instances can be passed to functions from other modules so 
+position_paradigm is the most important module in cgt. PositionParadigmFramework keeps track of important objects, can
+convert group elements and genomes between different forms, and instances can be passed to functions from other modules so
 that they can acces information without it needing to be recreated.
 """
 
-from sage.all_cmdline import *
+from sage.all_cmdline import QQ, FormalSum, SymmetricGroup, SymmetricGroupRepresentations, gap, UniversalCyclotomicField, Permutations, matrix, QQbar
 from sage.combinat.colored_permutations import SignedPermutations
 from .enums import *
 import numpy as np
 import warnings
-from copy import deepcopy 
+from copy import deepcopy
 from .structures import HyperoctahedralGroup
 from scipy.sparse import dok_matrix as dok
 from random import choice
+from functools import cache, lru_cache
 
 class PositionParadigmFramework:
     """Everything you need for working with genomes under the position paradigm"""
@@ -25,7 +26,7 @@ class PositionParadigmFramework:
         self.oriented = oriented
         self.symmetry = symmetry
 
-    def __eq__(self, other):
+    def __eq__(self, other): # String representation is unique and repr calls str
         return self.__repr__() == other.__repr__()
 
     def __str__(self):
@@ -34,12 +35,15 @@ class PositionParadigmFramework:
         return string
 
     def __repr__(self):
-        return self.__str__()
+        return self.__str__() # String representation is unique
 
     def __call__(self, x):
         """Return an object as a group algebra element if possible"""
         return self.genome_algebra()(self.cycles(x))
 
+    def __hash__(self):
+        return self.__str__().__hash__() # String rep is unique, so just hash that!
+
     def cycles(self, element):
         try: # See if it's already a group element
             return self.genome_group()(element)
@@ -66,8 +70,8 @@ class PositionParadigmFramework:
         Args:
             element: the genome instance to represent in one-row notation. Multiple formats accepted.
             as_list: if true, return as a list of images, otherwise return a sage permutation.
-        
-        TESTS::
+
+        EXAMPLES::
             sage: import cgt
             sage: ppf = cgt.PositionParadigmFramework(3)
             sage: ppf.cycles(ppf.one_row('(1,-2)(-1,2)')) == ppf.cycles('(1,-2)(-1,2)')
@@ -82,20 +86,17 @@ class PositionParadigmFramework:
                 row = Permutations(self.n)(row)
         return row
 
+    @cache
     def genome_group(self):
         """Return the permutation group containing genome instances."""
-        try:
-            return self.G
-        except AttributeError:
-            if self.oriented:
-                self.G = HyperoctahedralGroup(self.n)
-            else:
-                self.G = SymmetricGroup(self.n)
-        return self.G
+        if self.oriented:
+            return HyperoctahedralGroup(self.n)
+        else:
+            return SymmetricGroup(self.n)
 
     def canonical_instance(self, instance):
         """Return the 'canonical' instance of the genome represented by the permutation if there is one.
-        
+
         TESTS::
             sage: import cgt; ppf = cgt.PositionParadigmFramework(3, symmetry=cgt.SYMMETRY.circular)
             sage: ppf.canonical_instance('(1,-2)(-1,2)')
@@ -124,28 +125,22 @@ class PositionParadigmFramework:
         """Return a random genome"""
         return self.genome(self.random_instance(), format=format)
 
+    @cache
     def symmetry_group(self):
         """Return the symmetry group of the genomes."""
-        try:
-            return self.Z
-        except AttributeError:
-            if self.symmetry == SYMMETRY.circular:
-                gens = [self.standard_reflection(), self.standard_rotation()]
-            elif self.symmetry == SYMMETRY.linear:
-                gens = [self.standard_reflection()]
-            else:
-                gens = [self.genome_group().one()]
-            self.Z = self.G.subgroup(gens)
-        return self.Z
-    
+        if self.symmetry == SYMMETRY.circular:
+            gens = [self.standard_reflection(), self.standard_rotation()]
+        elif self.symmetry == SYMMETRY.linear:
+            gens = [self.standard_reflection()]
+        else:
+            gens = [self.genome_group().one()]
+        return self.genome_group().subgroup(gens)
+
+    @cache
     def group_algebra(self):
         """Return the group alegbra, where the group is the group containing genome instances."""
-        try:
-            return self.CG
-        except AttributeError:
-            self.GA = self.genome_group().algebra(QQ)
-        return self.GA
-        
+        return self.genome_group().algebra(QQ)
+
     def symmetry_element(self):
         """Return the symmetry element: a convex sum of elements from symmetry_group()."""
         try:
@@ -157,7 +152,7 @@ class PositionParadigmFramework:
     def num_genomes(self):
         """Return the number of distinct genomes up to symmetries."""
         return self.genome_group().order()/self.symmetry_group().order()
-    
+
     def _sort_key(self, one_row_perm):
         return str(one_row_perm).replace('-', 'Z')
 
@@ -226,7 +221,7 @@ class PositionParadigmFramework:
             for i in range(1, int(self.n/2)):
                 string += f'({i+1},{self.n-(i-1)-1})'
         return self.genome_group()(string)
-        
+
     def standard_rotation(self):
         """Return a permutation which rotates positions clockwise by one position."""
         positive = tuple(i for i in range(1,self.n+1))
@@ -238,6 +233,7 @@ class PositionParadigmFramework:
         return self.genome_group()(string)
 
     def draw_instance(self, instance, shortened=False):
+        """Return a one-line drawing of a genome using arrows indicating orientation and elipsis indicating rotational symmetry"""
         permutation = instance.inverse()
         if shortened:
             left_tail, right_head = '\u2758', '\u276D'
@@ -307,43 +303,53 @@ class PositionParadigmFramework:
         print('...done!')
         return matrix
 
-
     def irreps(self, element=None):
-        """Return a complete list of pairwise irreducible representations of the genome group."""
-        try:
-            representations = self.irreducible_representations
-        except AttributeError:
-            representations = []
-            def irrep_function_factory(irrep, signed):
-                def representation(sigma, _irrep=irrep, _signed=signed):
-                        result = 0
-                        if sigma in self.group_algebra(): # sigma is an algebra element
-                            for term in sigma:
-                                perm, coeff = term
-                                result += coeff * (gap.Image(_irrep, perm) if _signed else _irrep(perm))
-                        else: # sigma is a group element
-                            result = (gap.Image(_irrep, sigma) if _signed else _irrep(sigma))
-                        return matrix(UniversalCyclotomicField(), result)
-                return representation
-            if not self.oriented:
-                irreps = SymmetricGroupRepresentations(self.n)
-            else:
-                irreps = (gap.IrreducibleAffordingRepresentation(character) for character in gap.Irr(self.genome_group()))
-            for irrep in irreps:
-                representations.append(irrep_function_factory(irrep, self.oriented))
-            self.irreducible_representations = representations
+        """
+        Return a complete list of pairwise irreducible representations of the genome group.
+
+        Args:
+            element: return the image of element for each irreducible representation.
+
+        Returns:
+            a complete list of pairwise irreducible representations of the genome group, or a list of matrices if element is not None. 
+            Output is cached so that irreps are computed only once per session.
+            Matrix entries sit inside the UniversalCyclotomicField and are exact expressions.
+        """
+        representations = self._cached_irreps()
         if element is not None:
             return [irrep(element) for irrep in representations]
         else:
             return representations
 
-    def regular_representation(self, g): 
+    @cache
+    def _cached_irreps(self):
+        representations = []
+        def irrep_function_factory(irrep, signed):
+            def representation(sigma, _irrep=irrep, _signed=signed):
+                    result = 0
+                    if sigma in self.group_algebra(): # sigma is an algebra element
+                        for term in sigma:
+                            perm, coeff = term
+                            result += coeff * (gap.Image(_irrep, perm) if _signed else _irrep(perm))
+                    else: # sigma is a group element
+                        result = (gap.Image(_irrep, sigma) if _signed else _irrep(sigma))
+                    return matrix(UniversalCyclotomicField(), result)
+            return representation
+        if not self.oriented:
+            irreps = SymmetricGroupRepresentations(self.n)
+        else:
+            irreps = (gap.IrreducibleAffordingRepresentation(character) for character in gap.Irr(self.genome_group()))
+        for irrep in irreps:
+            representations.append(irrep_function_factory(irrep, self.oriented))
+        return representations
+
+    def regular_representation(self, g):
         """Return the regular representation of a single element"""
         warnings.warn("this function is untested! Use at your own risk.", DeprecationWarning)
         return matrix(QQbar, [(self.group_algebra()(g)*self.genome_group()(h)).to_vector(QQbar) for h in self.genome_group()]).transpose()
 
     def coefficient_in(self, more_terms, fewer_terms):
-        """For example, the coefficient of (a/2 + b/2) in x=(a/3 + b/3 + c/3) is 2/3, since x=2/3*(a/2 + b/2) + c/3""" 
+        """For example, the coefficient of (a/2 + b/2) in x=(a/3 + b/3 + c/3) is 2/3, since x=2/3*(a/2 + b/2) + c/3"""
         coefficients_in_larger_sum = {}
         for term in fewer_terms.terms():
             perm  = self.genome_group()(list(term)[0][0]) # a
@@ -355,6 +361,7 @@ class PositionParadigmFramework:
         return min(terms)
 
     def collect_genome_terms(self, formal_sum, display=DISPLAY.one_row):
+        """Return a weighted sum of genomes, represented as [canonical representation]z"""
         with warnings.catch_warnings():
             warnings.simplefilter('ignore')
             formal_sum = deepcopy(formal_sum)
